/*** Marcin Krzewski
/*** Grupa: N1-IBERE2

/*** Boost::Interprocess::SharedMemory
/*** Przykad pokazujcy uycie pamici dzielonej do wymiany danych pomidzy wtkami

Pami dzielona jest najszybszym sposobem komunikacji midzyprocesowej. System operacyjny mapuje
segment pamici w przestrzeni adresowej szeregu procesw, dziki czemu wiele procesw moe czyta 
oraz zapisywa ten segment pamici bez potrzeby woania funkcji systemu operacyjnego. Jednake, 
potrzebny jest sposb synchronizacji pomidzy procesami, ktre operuj na tym samym obszarze pamici
dzielonej. Dziki zastosowaniu pamici dzielonej, moemy unikn tworzenia dwch kopii danych w systemie.

Pami dzielona moe by wykorzystywana przez rne procesy, wtki pochodzce od jednego procesu oraz
wtki, ktre zostay stworzone w rnych procesach.

W pamici dzielonej moemy przechowywa podstawowe typy danych, jak rwnie typy i obiekty
zdefiniowane przez programist. ***/

// Nagwki umoliwiajce korzystanie z pamici dzielonej
#include <boost/interprocess/shared_memory_object.hpp>
#include <boost/interprocess/managed_shared_memory.hpp>

#include <boost/thread.hpp>
#include <iostream>

// Przestrze nazw komunikacji midzyprocesowej, w tym pamici dzielonej
using namespace boost::interprocess;
using namespace std;

// Wykonywanie operacji na danych z pamici dzielonej
void shmem(const int offset_) {
	try {
		// Otwieranie istniejcego obszaru pamici dzielonej
		shared_memory_object shm (
			open_only,			// otwieranie istniejcego bloku
			"shared_memory",	// nawa bloku
			read_write);		// z opcjami odczytu/zapisu

		// Mapowanie pamici dzielonej
		mapped_region region(shm, read_write);

		// Pobieranie adresy do mapowanego obszaru pamici
		void* addr = region.get_address();

		// Rzutowanie wskanika adresu na typu int*
		int* data = static_cast<int*>(addr);

		// Obliczanie nowej wartoci danej oraz umieszczenie jej z powrotem w pamici dzielonej
		data[offset_] *= data[offset_];
	}
	catch(interprocess_exception &ex) {
		std::cout << "Unexpected exception: " << ex.what() << std::endl;
		shared_memory_object::remove("shared_memory");
	}
}

void namedShm(const char* name) {
	try{
		// Specjalna pami dzielona, gdzie moemy konstruowa obiekty skojarzone z nazwami
		managed_shared_memory segment(open_only, "shared_memory");

		// Szukanie odpowiedniego obiektu w pamici
		// Funkcja managed_shared_memory::find zwraca kontener std::pair, gdzie pierwsz
		// wartoci jest wskanik do danych zawartych w pamici, a drugim wielko 
		// zarezerwowanej pamici
		pair<int*, size_t> res1 = segment.find<int>(name);
		if(res1.second != 0) {
			int* ptr = res1.first;
			for(unsigned int i = 0; i < res1.second; i++)
				cout << ptr[i] << ' ';

			cout << endl;
			return;
		}

		// Szukanie odpowiedniego obiektu w pamici
		pair<int*, size_t> res2 = segment.find<int>(name);   
		if(res2.second != 0) {
			int* ptr = res2.first;
			for(unsigned int i = 0; i < res2.second; i++)
				cout << ptr[i] << ' ';

			cout << endl;
			return;
		}
	}
	catch(interprocess_exception &ex) {
		std::cout << "Unexpected exception: " << ex.what() << std::endl;
		shared_memory_object::remove("shared_memory");
	}
}

int main ()
{
	// Przykad pokazuje powierzanie wykonywania operacji dodatkowym wtkom na danych umieszczonych 
	// w pamici dzielonej przez wtek gwny.
	try {
		// Usuwanie poprzedniej pamici dzielonej
		shared_memory_object::remove("shared_memory");

		// Tworzenie obiektu pamici dzielonej
		shared_memory_object shm(create_only,	// tworzenie bloku
							"shared_memory",	// nazwa bloku pamici
							read_write);		// opcje dostpu odczyt/zapis

		// Ustawienie wielkoci pamici
		shm.truncate(3);

		// Mapowanie caej pamici dzielonej dla tego procesu
		mapped_region region(shm, read_write);
		int* mem = static_cast<int*>(region.get_address());

		// Zapis rnych wartoci do pamici
		for(unsigned int i = 0; i < region.get_size(); i++) {
			mem[i] = i * 10 + 5;
			cout << mem[i] << ' ';
		}
		cout << endl;

		// Tworzenie nowych wtkw, ktre bd operoway na danych zawartych w pamici dzielonej
		boost::thread thread1(shmem, 0);
		boost::thread thread2(shmem, 1);
		boost::thread thread3(shmem, 2);

		// Oczekiwanie na zakoczenie poszczeglnych wtkw
		thread1.join();
		thread2.join();
		thread3.join();

		// Wywietlanie wartoci zmienionej przez wtki
		cout << "Obliczone wartosci: ";
		cout << mem[0] << ' ' << mem[1] << ' ' << mem[2] << endl;

		// Usuwanie pamici dzielone
		shared_memory_object::remove("shared_memory");
	}
	catch(interprocess_exception &ex) {
		shared_memory_object::remove("shared_memory");
		cout << ex.what() << std::endl;
		return 1;
	}

	// Przykad pokazuje uycie specjalnej pamici dzielonej, w ktrej moemy do skonstruowanego obiektu
	// przypisa nazw.
	try {
		// Usuwanie starej pamici dzielonej o takiej samej nazwie, tworzenie nowego bloku
		// oraz zarezerwowanie wymaganych zasobw
		shared_memory_object::remove("shared_memory");
		managed_shared_memory segment(create_only, "shared_memory", 65536);

		// Tworzy tablice 4 elementw typu int oraz inicjalizuje wszystkie wartoci 255
		int *array_int = segment.construct<int>("first int array")[4](255);

		// Tworzy tablice 3 elementw typu int oraz inicjalizuje kolejne z nich, 
		// wartoci zwracan przez wyraenie *tab_int++
		int tab_int[3] = {12, 34, 56};
		int *array_int2 = segment.construct_it<int>("second int array")[3](&tab_int[0]); 

		boost::thread thread1(namedShm, "first int array");
		boost::thread thread2(namedShm, "second int array");
		thread1.join();
		thread2.join();
		shared_memory_object::remove("shared_memory");
	}
	catch(interprocess_exception &ex) {
		std::cout << "Unexpected exception: " << ex.what() << std::endl;
		shared_memory_object::remove("shared_memory");
	}

	getchar();
	return 0;
}